package org.netkernel.lang.dpml.endpoint;

import org.netkernel.container.ILogger;
import org.netkernel.lang.dpml.ast.*;
import org.netkernel.lang.dpml.ast.impl.LocationImpl;
import org.netkernel.lang.dpml.ast.impl.LiteralIndentifier;
import org.netkernel.lang.dpml.ast.impl.LiteralValue;
import org.netkernel.lang.dpml.ast.impl.Closure;
import org.netkernel.lang.dpml.representation.DPMLRepresentation;
import org.netkernel.layer0.bnf.GrammarParseException;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFRequestImpl;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.urii.OrExpiryFunction;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IExpiry;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.*;
import org.netkernel.urii.ISpace;


public class DPMLRuntime extends StandardAccessorImpl
{
	private static boolean sShouldLog;
	private static DPMLRuntime sInstance;
	
	public DPMLRuntime() throws GrammarParseException
	{	declareThreadSafe();
		declareSupportedVerbs(INKFRequestReadOnly.VERB_SOURCE|INKFRequestReadOnly.VERB_SINK|INKFRequestReadOnly.VERB_EXISTS|INKFRequestReadOnly.VERB_DELETE|INKFRequestReadOnly.VERB_NEW);
		declareVerboseUnhandledExceptions(false);
	}
	
	public static boolean shouldLog()
	{	return sShouldLog;
	}
	
	public static DPMLRuntime getInstance()
	{	return sInstance;
	}

	public void onRequest(INKFRequestContext aContext) throws Exception
	{
		sInstance=this;
		if (aContext.getThisRequest().getVerb()==INKFRequestReadOnly.VERB_META)
		{	super.onMeta(aContext);
		}
		else
		{
			// get the script
			String scriptIdentifier = aContext.getThisRequest().getArgumentValue("operator");
			ILocation loc = new LocationImpl(scriptIdentifier,-1,-1);
			aContext.setCWU(scriptIdentifier);
			INKFResponseReadOnly<DPMLRepresentation> scriptResponse=aContext.sourceForResponse(scriptIdentifier, DPMLRepresentation.class);
			DPMLRepresentation script = scriptResponse.getRepresentation();
			
			IRequestScopeLevel rctx = aContext.getKernelContext().getThisKernelRequest().getRequestScope();
			
			//put all of our arguments into the context
			DPMLSubrequestContext argContext = new DPMLSubrequestContext(rctx,getSpace(),scriptIdentifier);
			rctx = RequestScopeLevelImpl.createDurableScopeLevel(argContext,rctx);
			
			INKFRequestReadOnly thisReq=aContext.getThisRequest();
			for (int i=thisReq.getArgumentCount()-1; i>=0; i--)
			{	String name=thisReq.getArgumentName(i);
				String reference=thisReq.getArgumentValue(i);
				IVariable var = new LiteralIndentifier(name,null,loc,reference,true);
				argContext.putArgument(name,var);
			}
					
			//add verb
			String verbString = RequestFactory.verbString( aContext.getThisRequest().getVerb() );
			if (verbString==null)
			{	throw aContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_VERB",null,null,null);
			}
			RequestResponseFieldsImpl userMeta = new RequestResponseFieldsImpl(1);
			userMeta.put(MetaImpl.KEY_MIMETYPE,"text/plain");
			MetaImpl meta = new MetaImpl(ExpiryFactory.getNeverExpired(),userMeta);
			argContext.putArgument(IArgument.ARG_VERB,new LiteralValue(IArgument.ARG_VERB,null,loc,verbString,meta));
			
			argContext.onCommissionSpace(aContext.getKernelContext().getKernel());
	
			//put all the static closures into the context
			ISpace scriptContext = script.getScriptContext();
			rctx = RequestScopeLevelImpl.createDurableScopeLevel(scriptContext,rctx);
			
			//logging
			ILogger logger=aContext.getKernelContext().getKernel().getLogger();
			sShouldLog=logger.shouldLog(ILogger.DEBUG,this);
			
			if (sShouldLog)
			{	
				StringBuffer sb=new StringBuffer(64);
				for (int i=0; i<thisReq.getArgumentCount(); i++)
				{	if (i!=0)
					{	sb.append(", ");
					}
					String name = thisReq.getArgumentName(i);
					sb.append(name);
				}
							
				Object[] args = new Object[] { ((IVariable)script.getRootScope()).getLocation().getSourceIdentity(), verbString, sb.toString() };
				String msg=aContext.formatMessage("MSG_LOG_ENTER",args);
				logger.logRaw(ILogger.DEBUG,this,msg);
			}
			
			IVariable rootResponse = script.getRootScope();
			if (aContext.getThisRequest().argumentExists("response"))
			{	String responseClosure=aContext.getThisRequest().getArgumentValue("response");
				if (rootResponse instanceof Closure)
				{	rootResponse=((Closure) rootResponse).resolveVariable(responseClosure, rctx);
					if (rootResponse==null)
					{	throw aContext.createFormattedException("EX_DPML_RUNTIME","MSG_RUN_NOT_FOUND",null,null,responseClosure);
					}
				}
				else
				{	throw aContext.createFormattedException("EX_DPML_RUNTIME","MSG_ARG_RESPONSE_NOT_CLOSURE",null,null);
				}
			}
			
			
			//IVariable rootResponse=script.getRootScope();
			
			INKFResponse response;
			try
			{	response = processRequest3(aContext,rootResponse,rctx,scriptResponse);
			}
			catch (NKFException e)
			{	Throwable t=e;
				if (e.getId().equals(NKFException.class.getName()))
				{	t=e.getCause();
				}
				response=aContext.createResponseFrom(t);
			}
			
			if (sShouldLog)
			{	String repString=null;
				if (response!=null && response.getRepresentation()!=null)
				{	repString=response.getRepresentation().toString();
				}
				Object[] args = new Object[] { ((IVariable)script.getRootScope()).getLocation().getSourceIdentity(), repString };
				String msg=aContext.formatMessage("MSG_LOG_EXIT",args);
				logger.logRaw(ILogger.DEBUG,this,msg);
			}
		}
	}
	
	/**
	 * @param aContext
	 * @param aResponse
	 * @param aSubRequestContext
	 * @param aScriptResponse the response from the script request passed in only for the handling of the complete script (used to create a combined expiry function)
	 * @return
	 * @throws Exception
	 */
	public static INKFResponse processRequest3(INKFRequestContext aContext, IVariable aResponse, IRequestScopeLevel aSubRequestContext, INKFResponseReadOnly aScriptResponse) throws Exception
	{
		IVariable var=dereference(aResponse,aContext,aSubRequestContext);
		INKFResponse result;
		
		if (var instanceof IIdentifierVariable)
		{	IIdentifierVariable idVar = (IIdentifierVariable)var;
			String id=idVar.getIdentifier(aContext,aSubRequestContext);
			INKFRequest req=aContext.createRequest(id);
			((NKFRequestImpl)req).setRequestScope(aSubRequestContext);
			INKFResponseReadOnly resp=aContext.issueRequestForResponse(req);
			result=aContext.createResponseFrom(resp);
		}
		else if (var instanceof IValueVariable)
		{	IValueVariable valueVar = (IValueVariable)var;
			IValue value=valueVar.getValue(aContext,aSubRequestContext,aContext.getThisRequest().getRepresentationClass());
			if (value!=null)
			{	result=createResponseWithMetaData(value,aContext);
				IRequestResponseFields responseHeaders=value.getMeta().getUserMetaData();
				if (aScriptResponse!=null && responseHeaders.getValue(ResponseAccessor.HEADER_EXPIRY)!=null)
				{	setResponseExpiry(result,responseHeaders,aScriptResponse);
				}
			}
			else
			{	result=null;
			}
		}
		else
		{	System.out.println("found null");
			throw new NullPointerException();
		}		
		return result;
	}
	
	public static IVariable dereference(IVariable aVar, INKFRequestContext aContext, IRequestScopeLevel aSubRequestContext) throws NKFException
	{
		IVariable result=aVar;
		String name=null;
		IVariable lastDeref=result;
		
		DPMLSubrequestContext ctx=null;//(aSubRequestContext.getSpace() instanceof DPMLSubrequestContext)?(DPMLSubrequestContext)aSubRequestContext.getSpace():null;
		
		while(result instanceof IReferenceVariable)
		{	lastDeref=result;
			name=((IReferenceVariable)result).getReference();
			boolean set=false;
			
			if (ctx!=null)
			{	IVariable v =ctx.resolveName(name);
				if (v!=null)
				{	result=v;
					set=true;
				}
			}
			if (!set)
			{	result=resolveName(name,result.getScope(), aSubRequestContext);
			}
		}
		if (result==null)
		{	throw aContext.createFormattedException("EX_DPML_RUNTIME","MSG_RUN_NOT_FOUND",formatLocationMessage(lastDeref,aContext),null,name);
		}
		return result;
	}
	
	public static String formatLocationMessage(IVariable aVar, INKFRequestContext aContext)
	{	ILocation loc = aVar.getLocation();
		if (loc!=null)
		{	return aContext.formatMessage("MSG_PXML_LOCATION",new Object[] {loc.getSourceIdentity(),new Integer(loc.getLine()),new Integer(loc.getColumn())});
		}
		else
		{	return null;
		}
	}
	
	
	public static IVariable resolveName(String aName, IScope aScope, IRequestScopeLevel aDynamicScope)
	{	IScope scope=aScope;
		IVariable result = null;
		while(scope!=null && result==null)
		{	result=scope.resolveVariable(aName, aDynamicScope);
			scope=scope.getParent();
		}
		return result;
	}
	
	/** take the argument and apply the method if necessary to modify the variable
	 * @param aArg
	 * @param aContext
	 * @param aSubRequestContext
	 * @return
	 * @throws NKFException
	 */
	public static IVariable processArgument(IArgument aArg, INKFRequestContext aContext, IRequestScopeLevel aSubRequestContext, boolean aAlwaysDereference) throws NKFException
	{	
		IVariable inVar=aArg.getVariable();
		ILocation inLoc=inVar.getLocation();
		String inName=inVar.getName();
		
		if (aAlwaysDereference || aArg.getMethod()==IArgument.METHOD_IDENTITY || aArg.getMethod()==IArgument.METHOD_EAGER)
		{	inVar=dereference(inVar,aContext,aSubRequestContext);
		}
		IVariable result=inVar;
		boolean isPrimary=aArg.getName().equals(IArgument.ARG_PRIMARY);
		
		if (!(inVar instanceof IReferenceVariable))
		{
			if (aArg.getMethod()==IArgument.METHOD_EAGER || isPrimary)
			{	
				if (inVar instanceof IIdentifierVariable)
				{	//source identifier
					String identifier = ((IIdentifierVariable)inVar).getIdentifier(aContext,aSubRequestContext);
					INKFRequest subReq=aContext.createRequest(identifier);
					((NKFRequestImpl)subReq).setRequestScope(aSubRequestContext);
					INKFResponseReadOnly resp=aContext.issueRequestForResponse(subReq);
					result=new LiteralValue(inName,null,inLoc,resp.getRepresentation(),((NKFResponseReadOnlyImpl)resp).getKernelResponse().getMeta());					
				}
				else // IValueVariable
				{	IValue value = ((IValueVariable)inVar).getValue(aContext,aSubRequestContext,Object.class);
					result=new LiteralValue(inName,null,inLoc,value.getRepresentation(),value.getMeta());							
				}
			}
			else if (aArg.getMethod()==IArgument.METHOD_IDENTITY)
			{	
				INKFRequest subReq=aContext.createRequest("active:toDataURI");
				if (inVar instanceof IIdentifierVariable)
				{	String identifier = ((IIdentifierVariable)inVar).getIdentifier(aContext,aSubRequestContext);
					subReq.addArgument("operand", identifier);
				}
				else // IValueVariable
				{	IValue value = ((IValueVariable)inVar).getValue(aContext,aSubRequestContext,IBinaryStreamRepresentation.class);
					INKFResponse valResp = createResponseWithMetaData(value,aContext);
					subReq.addArgumentFromResponse("operand", valResp);
				}
				subReq.setRepresentationClass(String.class);
				String rep = (String)aContext.issueRequest(subReq);
				result = new LiteralIndentifier(inName,null,inLoc,rep,true);
			}
			else
			{	
				if (inVar instanceof IIdentifierVariable)
				{	String identifier = ((IIdentifierVariable)inVar).getIdentifier(aContext,aSubRequestContext);
					if (identifier.startsWith(DPMLSubrequestContext.ARG_SCHEME_ESC))
					{	String name=identifier.substring(DPMLSubrequestContext.ARG_SCHEME_ESC.length());
						//corner case of an identifier to an arg, see if this has an identifier origin and if so, use that identifier directly
						// (allows argument identifiers to pass through dpml unchanged)
						IRequestScopeLevel scope=aSubRequestContext;
						while(scope!=null)
						{	ISpace s=scope.getSpace();
							if (s instanceof DPMLSubrequestContext)
							{	DPMLSubrequestContext dsc = (DPMLSubrequestContext)s;
								IVariable v=dsc.resolveName(name);
								if (v instanceof IIdentifierVariable)
								{	String argId=((IIdentifierVariable)v).getIdentifier(aContext, dsc.getContextForArguments());
									result=v;
									break;
								}
							}
							scope=scope.getParent();
						}
					}
				}
			}
		}
		return result;
	}
	
	public static Object ensureClassOfValue(IValue aValue, Class aRepresentationClass, INKFRequestContext aContext) throws NKFException
	{
		Object result;
		if (aRepresentationClass.isAssignableFrom(aValue.getRepresentation().getClass()))
		{	result=aValue.getRepresentation();
		}
		else
		{	INKFResponse valResp = createResponseWithMetaData(aValue,aContext);
			String id=SimpleIdentifierImpl.getUnique("literal:var").toString();
			INKFRequest subReq=aContext.createRequest(id);
			subReq.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
			subReq.addPrimaryArgumentFromResponse(valResp);
			subReq.setRepresentationClass(aRepresentationClass);
			result=aContext.issueRequest(subReq);
		}
		return result;
	}
	
	private static void setResponseExpiry(INKFResponse aResponse, IRequestResponseFields aResponseHeaders, INKFResponseReadOnly aScriptResponse)
	{
		Integer modifiedExpiry=(Integer)aResponseHeaders.getValue(ResponseAccessor.HEADER_EXPIRY);
		Long modifiedExpiryPeriod=(Long)aResponseHeaders.getValue(ResponseAccessor.HEADER_EXPIRY_PERIOD);
		final IExpiry scriptExpiry=((NKFResponseReadOnlyImpl)aScriptResponse).getMeta().getExpiry();
		
		switch(modifiedExpiry)
		{	case INKFResponse.EXPIRY_ALWAYS:
				aResponse.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				break;
			case INKFResponse.EXPIRY_NEVER:
				aResponse.setExpiry(INKFResponse.EXPIRY_NEVER);
				break;
			case INKFResponse.EXPIRY_DEPENDENT:
				aResponse.setExpiry(INKFResponse.EXPIRY_DEPENDENT);
				break;
			case INKFResponse.EXPIRY_CONSTANT:
				aResponse.setExpiry(INKFResponse.EXPIRY_CONSTANT,modifiedExpiryPeriod+System.currentTimeMillis());
				break;
			case INKFResponse.EXPIRY_MAX_CONSTANT_DEPENDENT:
				aResponse.setExpiry(INKFResponse.EXPIRY_MAX_CONSTANT_DEPENDENT,modifiedExpiryPeriod+System.currentTimeMillis());
				break;
			case INKFResponse.EXPIRY_MIN_CONSTANT_DEPENDENT:
				aResponse.setExpiry(INKFResponse.EXPIRY_MIN_CONSTANT_DEPENDENT,modifiedExpiryPeriod+System.currentTimeMillis());
				break;
			default:
				//invalid value set
		}
	}
	
	public static INKFResponse createResponseWithMetaData(IValue aValue, INKFRequestContext aContext)
	{	INKFResponse result= aContext.createResponseFrom(aValue.getRepresentation());
		IResponseMeta meta=aValue.getMeta();
		IRequestResponseFields headers = meta.getUserMetaData();
		if (headers.getValue(ResponseAccessor.HEADER_EXPIRY)!=null)
		{	//need to remove this header to stop it propagating up and effecting someone else
			int s=headers.size();
			for (int i=0; i<s; i++)
			{	String key=headers.getKey(i);
				if (!key.equals(ResponseAccessor.HEADER_EXPIRY))
				{	result.setHeader(key, headers.getValue(i));
				}
			}
		}
		else
		{	result.setHeaders(headers);
		}
		return result;
	}
}